c***********************************************************************
c
c  File:        lsmdd_patchmodule_fort.f
c  Copyright:   (c) 2005-2007 Princeton University
c  Author(s):   Kevin T. Chu
c  Revision:    $Revision: 376 $
c  Modified:    $Date: 2007-08-06 19:10:55 -0400 (Mon, 06 Aug 2007) $
c  Description: F77 patch module routines for test problems involving 
c               obstacles
c
c***********************************************************************

c***********************************************************************
c  initializeStraightLineDislocation() initializes phi and psi for an 
c  arbitrary straight-line dislocation with tangent vector parallel
c  to one of the coordinate axis.
c
c  Arguments:
c    phi (out):               phi data
c    psi (out):               psi data
c    *_gb (in):               ghostbox for phi and psi data
c    *_fb (in):               fillbox for phi and psi data
c    x_lower (in):            physical coordinates of lower corner of box
c    dx (in):                 grid spacing
c    domain_dimensions (in):  physical dimensions of domain
c    pt_* (in):               coordinate of a point on the line
c    dir_* (in):              components of tangent vector 
c                             (not necessarily normalized)
c
c***********************************************************************
      subroutine initializeStraightLineDislocation(
     &  phi,
     &  ilo_phi_gb, ihi_phi_gb, 
     &  jlo_phi_gb, jhi_phi_gb, 
     &  klo_phi_gb, khi_phi_gb,
     &  psi,
     &  ilo_psi_gb, ihi_psi_gb, 
     &  jlo_psi_gb, jhi_psi_gb, 
     &  klo_psi_gb, khi_psi_gb,
     &  ilo_fb, ihi_fb, jlo_fb, jhi_fb, klo_fb, khi_fb,
     &  x_lower, 
     &  dx,
     &  domain_dimensions,
     &  pt_x, pt_y, pt_z,
     &  dir_x, dir_y, dir_z)
c***********************************************************************
      implicit none

c     _gb refers to ghost box
c     _fb refers to fill box
      integer ilo_phi_gb, ihi_phi_gb
      integer jlo_phi_gb, jhi_phi_gb
      integer klo_phi_gb, khi_phi_gb
      integer ilo_psi_gb, ihi_psi_gb
      integer jlo_psi_gb, jhi_psi_gb
      integer klo_psi_gb, khi_psi_gb
      integer ilo_fb, ihi_fb, jlo_fb, jhi_fb, klo_fb, khi_fb
      double precision phi(ilo_phi_gb:ihi_phi_gb,
     &                     jlo_phi_gb:jhi_phi_gb,
     &                     klo_phi_gb:khi_phi_gb)
      double precision psi(ilo_psi_gb:ihi_psi_gb,
     &                     jlo_psi_gb:jhi_psi_gb,
     &                     klo_psi_gb:khi_psi_gb)
      double precision x_lower(1:3)
      double precision dx(1:3)
      double precision domain_dimensions(1:3)
      double precision pt_x, pt_y, pt_z
      double precision dir_x, dir_y, dir_z
      integer i,j,k
      double precision pos_minus_pt_x, pos_minus_pt_y, pos_minus_pt_z
      double precision n1_x, n1_y, n1_z
      double precision n2_x, n2_y, n2_z
      double precision norm_tangent_vector
      double precision pos_minus_pt_cross_tan_x
      double precision pos_minus_pt_cross_tan_y
      double precision pos_minus_pt_cross_tan_z
      double precision r
      double precision theta
      double precision r_cos_theta, r_sin_theta
      double precision norm_n1
      double precision norm_n2
      double precision tol
      parameter (tol = 1.0d-8)

c     compute components of arbitrary vector normal to dislocation line
      if ( (abs(dir_y) .lt. tol) .and. (abs(dir_z) .lt. tol) ) then
c       use (0,1,0) and tangent vector to define plane
        n1_x = 0.0d0
        n1_y = 1.0d0
        n1_z = 0.0d0
        n2_x = 0.0d0
        n2_y = 0.0d0
        n2_z = sign(1.0d0,dir_x)
      else
c       use (1,0,0) and tangent vector to define plane
        norm_n1 = sqrt( dir_y**2 + dir_z**2 )
        n1_x = 0.0d0
        n1_y = dir_z/norm_n1
        n1_z = -dir_y/norm_n1
        n2_x = dir_y*n1_z - dir_z*n1_y
        n2_y = -dir_x*n1_z
        n2_z = dir_x*n1_y
        norm_n2 = sqrt(n2_x**2 + n2_y**2 + n2_z**2)
        n2_x = n2_x/norm_n2
        n2_y = n2_y/norm_n2
        n2_z = n2_z/norm_n2
      endif

c     compute norm of tangent vector
      norm_tangent_vector = sqrt( dir_x**2 + dir_y**2 + dir_z**2 )

c     loop over grid {
      do k=klo_fb,khi_fb
        do j=jlo_fb,jhi_fb
          do i=ilo_fb,ihi_fb

c           compute distance from point to dislocation line
c           passing through pt
            pos_minus_pt_x = x_lower(1) + dx(1)*(i-ilo_fb+0.5)
     &                     - pt_x
            pos_minus_pt_y = x_lower(2) + dx(2)*(j-jlo_fb+0.5)
     &                     - pt_y
            pos_minus_pt_z = x_lower(3) + dx(3)*(k-klo_fb+0.5)
     &                     - pt_z

            pos_minus_pt_cross_tan_x = 
     &        pos_minus_pt_y * dir_z - pos_minus_pt_z * dir_y
            pos_minus_pt_cross_tan_y =
     &        pos_minus_pt_z * dir_x - pos_minus_pt_x * dir_z
            pos_minus_pt_cross_tan_z =
     &        pos_minus_pt_x * dir_y - pos_minus_pt_y * dir_x
            r = sqrt( pos_minus_pt_cross_tan_x**2
     &              + pos_minus_pt_cross_tan_y**2
     &              + pos_minus_pt_cross_tan_z**2 ) 
     &        / norm_tangent_vector

c           compute angle formed by tangent vector and n1
            r_cos_theta = ( pos_minus_pt_x*n1_x 
     &                    + pos_minus_pt_y*n1_y
     &                    + pos_minus_pt_z*n1_z)
            r_sin_theta = ( pos_minus_pt_x*n2_x 
     &                    + pos_minus_pt_y*n2_y
     &                    + pos_minus_pt_z*n2_z)

            if (r .gt. tol) then
              theta = atan2(r_sin_theta, r_cos_theta)
            else
              theta = 0.0d0
            endif

            phi(i,j,k) = r*cos( theta )
            psi(i,j,k) = r*sin( theta )

          enddo
        enddo
      enddo
c     } end loop over grid 

      return
      end
c***********************************************************************
