c***********************************************************************
c
c  File:        lsmdd_initialization.f
c  Copyright:   (c) 2005-2007 Princeton University
c  Author(s):   Kevin T. Chu
c  Revision:    $Revision: 1.2 $
c  Modified:    $Date: 2006/05/29 15:06:01 $
c  Description: F77 routines for initializing level set functions
c               for common dislocation configurations
c
c***********************************************************************

c***********************************************************************
c  initializeCircularDislocationLoop() initializes phi and psi a 
c  circular dislocation loop with the specified position, orientation
c  and radius.
c
c  Arguments:
c    phi (out):               phi data
c    psi (out):               psi data
c    normal_* (in):           coordinates of the normal to the plane
c                             containing the dislocation loop
c    center_* (in):           coordinates of center of dislocation loop
c    radius (in):             radius of circular dislocation loop
c    *_gb (in):               ghostbox for phi and psi data
c    *_fb (in):               fillbox for phi and psi data
c    x_lower (in):            physical coordinates of lower corner of box
c    domain_dims (in):        dimensions of physical domain
c    dx (in):                 grid spacing
c
c NOTES:
c - The direction of the loop is such that it satisfies the right-hand 
c   rule with respect to the specified normal vector.
c
c***********************************************************************
      subroutine initializeCircularDislocationLoop(
     &  phi,
     &  ilo_phi_gb, ihi_phi_gb, 
     &  jlo_phi_gb, jhi_phi_gb, 
     &  klo_phi_gb, khi_phi_gb,
     &  psi,
     &  ilo_psi_gb, ihi_psi_gb, 
     &  jlo_psi_gb, jhi_psi_gb, 
     &  klo_psi_gb, khi_psi_gb,
     &  ilo_fb, ihi_fb, jlo_fb, jhi_fb, klo_fb, khi_fb,
     &  normal,
     &  center,
     &  radius,
     &  x_lower, 
     &  dx)
c***********************************************************************
c   { begin subroutine

      implicit none

c     _gb refers to ghost box
c     _fb refers to fill box
      integer ilo_phi_gb, ihi_phi_gb
      integer jlo_phi_gb, jhi_phi_gb
      integer klo_phi_gb, khi_phi_gb
      integer ilo_psi_gb, ihi_psi_gb
      integer jlo_psi_gb, jhi_psi_gb
      integer klo_psi_gb, khi_psi_gb
      integer ilo_fb, ihi_fb, jlo_fb, jhi_fb, klo_fb, khi_fb
      double precision phi(ilo_phi_gb:ihi_phi_gb,
     &                     jlo_phi_gb:jhi_phi_gb,
     &                     klo_phi_gb:khi_phi_gb)
      double precision psi(ilo_psi_gb:ihi_psi_gb,
     &                     jlo_psi_gb:jhi_psi_gb,
     &                     klo_psi_gb:khi_psi_gb)
      double precision normal(1:3)
      double precision unit_normal_x, unit_normal_y, unit_normal_z
      double precision center(1:3)
      double precision radius
      double precision x_lower(1:3)
      double precision dx(1:3)
      integer i,j,k
      double precision x,y,z
      double precision norm
      double precision pt_minus_center_x
      double precision pt_minus_center_y
      double precision pt_minus_center_z
      double precision proj_along_normal
      double precision proj_in_plane
      double precision tol
      parameter (tol = 1.0d-8)

c     normalize normal vector
      norm = sqrt(normal(1)**2 + normal(2)**2 + normal(3)**2)
      unit_normal_x = normal(1)/norm
      unit_normal_y = normal(2)/norm
      unit_normal_z = normal(3)/norm

c     loop over grid {
      do k=klo_fb,khi_fb
        do j=jlo_fb,jhi_fb
          do i=ilo_fb,ihi_fb

            x = x_lower(1) + dx(1)*(i-ilo_fb+0.5)
            y = x_lower(2) + dx(2)*(j-jlo_fb+0.5)
            z = x_lower(3) + dx(3)*(k-klo_fb+0.5)

            pt_minus_center_x = x-center(1)
            pt_minus_center_y = y-center(2)
            pt_minus_center_z = z-center(3)

            proj_along_normal = pt_minus_center_x*unit_normal_x
     &                        + pt_minus_center_y*unit_normal_y
     &                        + pt_minus_center_z*unit_normal_z

            proj_in_plane = (pt_minus_center_x
     &                      -proj_along_normal*unit_normal_x)**2
     &                    + (pt_minus_center_y
     &                      -proj_along_normal*unit_normal_y)**2
     &                    + (pt_minus_center_z
     &                      -proj_along_normal*unit_normal_z)**2
            proj_in_plane = sqrt(proj_in_plane)

            phi(i,j,k) = proj_in_plane - radius
            psi(i,j,k) = proj_along_normal

          enddo
        enddo
      enddo
c     } end loop over grid 

      return
      end
c   } end subroutine
c***********************************************************************

c***********************************************************************
c  initializeStraightDislocationLine() initializes phi and psi
c  for a straight dislocation line with the specified direction
c  and passing through the specified point.  
c
c  This function only supports dislocation line directions along one
c  of the three coordinate axes.  It may be used for other line    
c  directions, but there is no guarantee that it will lead to the    
c  desired results for arbitrary line directions.  In particular,
c  the boundary conditions may difficult to prescribe.
c
c  Arguments:
c    phi (out):         phi data
c    psi (out):         psi data
c    xi (in):           direction of dislocation line
c    pt (in):           coordinates of a point that lies on
c                       the dislocation line
c    *_gb (in):         ghostbox for phi and psi data
c    *_fb (in):         fillbox for phi and psi data
c    x_lower (in):      physical coordinates of lower corner of box
c    domain_dims (in):  dimensions of physical domain
c    dx (in):           grid spacing
c
c NOTES:
c  - Boundary conditions should be set as follows for a dislocation line
c    that is parallel to the x-axis:
c    - phi x: periodic
c    - phi y: periodic
c    - phi z: anti-periodic
c    - psi x: periodic
c    - psi y: anti-periodic
c    - psi z: periodic
c   
c    Boundary conditions for dislocation lines oriented along other 
c    axes are just cyclical shifts of the above.
c
c***********************************************************************
      subroutine initializeStraightDislocationLine(
     &  phi,
     &  ilo_phi_gb, ihi_phi_gb, 
     &  jlo_phi_gb, jhi_phi_gb, 
     &  klo_phi_gb, khi_phi_gb,
     &  psi,
     &  ilo_psi_gb, ihi_psi_gb, 
     &  jlo_psi_gb, jhi_psi_gb, 
     &  klo_psi_gb, khi_psi_gb,
     &  ilo_fb, ihi_fb, jlo_fb, jhi_fb, klo_fb, khi_fb,
     &  xi, pt,
     &  x_lower, 
     &  domain_dims,
     &  dx)
c***********************************************************************
c   { begin subroutine

      implicit none

c     _gb refers to ghost box
c     _fb refers to fill box
      integer ilo_phi_gb, ihi_phi_gb
      integer jlo_phi_gb, jhi_phi_gb
      integer klo_phi_gb, khi_phi_gb
      integer ilo_psi_gb, ihi_psi_gb
      integer jlo_psi_gb, jhi_psi_gb
      integer klo_psi_gb, khi_psi_gb
      integer ilo_fb, ihi_fb, jlo_fb, jhi_fb, klo_fb, khi_fb
      double precision phi(ilo_phi_gb:ihi_phi_gb,
     &                     jlo_phi_gb:jhi_phi_gb,
     &                     klo_phi_gb:khi_phi_gb)
      double precision psi(ilo_psi_gb:ihi_psi_gb,
     &                     jlo_psi_gb:jhi_psi_gb,
     &                     klo_psi_gb:khi_psi_gb)
      double precision xi(1:3)
      double precision pt(1:3)
      double precision x_lower(1:3)
      double precision domain_dims(1:3)
      double precision dx(1:3)
      double precision pt_image(1:3)

c     compute image point outside of computational domain
      if (xi(1) .ne. 0.0) then
        pt_image(1) = pt(1)
        pt_image(2) = pt(2)+domain_dims(2)
        pt_image(3) = pt(3)
      elseif (xi(2) .ne. 0.0) then
        pt_image(1) = pt(1)
        pt_image(2) = pt(2)
        pt_image(3) = pt(3)+domain_dims(3)
      else
        pt_image(1) = pt(1)+domain_dims(1)
        pt_image(2) = pt(2)
        pt_image(3) = pt(3)
      endif

      call initializeDislocationArray(
     &  phi,
     &  ilo_phi_gb, ihi_phi_gb, 
     &  jlo_phi_gb, jhi_phi_gb, 
     &  klo_phi_gb, khi_phi_gb,
     &  psi,                  
     &  ilo_psi_gb, ihi_psi_gb, 
     &  jlo_psi_gb, jhi_psi_gb, 
     &  klo_psi_gb, khi_psi_gb,
     &  ilo_fb, ihi_fb, jlo_fb, jhi_fb, klo_fb, khi_fb,
     &  xi,
     &  pt,
     &  pt_image,
     &  x_lower,
     &  domain_dims,
     &  dx)


      return
      end
c   } end subroutine
c***********************************************************************

c***********************************************************************
c  initializeDislocationArray() initializes phi and psi for a periodic
c  array of straight dislocations with the specified direction.  
c  The distance/period between adjacent dislocation lines is set by
c  specifying points on two immediately adjacent dislocation lines.
c  
c  This function only supports dislocation arrays that lie in one of
c  the three coordinate planes (i.e. xy, yz, xz).  It may be used for    
c  orientations of the dislocation array, but there is no guarantee    
c  that it will lead to the desired results for arbitrarily oriented    
c  dislocation arrays.  In particular, the boundary conditions may     
c  difficult to prescribe.
c
c  Arguments:
c    phi (out):               phi data
c    psi (out):               psi data
c    xi (in):                 direction of dislocation lines
c    pt1 (in):                coordinates of first point that 
c                             dislocation array passes through
c    pt2 (in):                coordinates of second point that 
c                             dislocation array passes through
c    *_gb (in):               ghostbox for phi and psi data
c    *_fb (in):               fillbox for phi and psi data
c    x_lower (in):            physical coordinates of lower corner of 
c                             patch
c    domain_dimensions (in):  physical dimensions of domain
c    dx (in):                 grid spacing
c
c
c  NOTES:
c  - It is the users responsibility to ensure that placing 
c    dislocation lines passing through pt1 and pt2 leads to a 
c    dislocation array that is periodic with respect to the 
c    simulation box.
c
c  - Boundary conditions should be set as follows.  Let N denote the
c    direction perpendicular to the plane of the dislocation array.
c    - if the line direction is parallel to one of the coordinate axes:
c      - phi: 
c        - periodic in the direction of xi 
c        - anti-periodic in the direction of N
c        - periodic in the direction of (xi cross N) if there are an 
c          even number of dislocation lines in the simulation cell;
c          anti-periodic otherwise
c      - psi: 
c        - periodic in the direction of xi and N
c        - periodic in the direction of (xi cross N) if there are an 
c          even number of dislocation lines in the simulation cell;
c          anti-periodic otherwise
c    - if the line direction is not parallel to any of the coordinate axes:
c      - phi:  
c        - anti-periodic in the direction of N
c        - anti-periodic in both directions that lie in the plane of the
c          dislocation array if there are an even number of dislocation
c          lines in the simulation cell; periodic otherwise.
c      - psi:  
c        - periodic in the direction of N
c        - anti-periodic in both directions that lie in the plane of the
c          dislocation array if there are an even number of dislocation
c          lines in the simulation cell; periodic otherwise.
c         
c
c***********************************************************************
      subroutine initializeDislocationArray(
     &  phi,
     &  ilo_phi_gb, ihi_phi_gb, 
     &  jlo_phi_gb, jhi_phi_gb, 
     &  klo_phi_gb, khi_phi_gb,
     &  psi,
     &  ilo_psi_gb, ihi_psi_gb, 
     &  jlo_psi_gb, jhi_psi_gb, 
     &  klo_psi_gb, khi_psi_gb,
     &  ilo_fb, ihi_fb, jlo_fb, jhi_fb, klo_fb, khi_fb,
     &  xi,
     &  pt1,
     &  pt2,
     &  x_lower, 
     &  domain_dimensions, 
     &  dx)
c***********************************************************************
c   { begin subroutine

      implicit none

c     _gb refers to ghost box
c     _fb refers to fill box
      integer ilo_phi_gb, ihi_phi_gb
      integer jlo_phi_gb, jhi_phi_gb
      integer klo_phi_gb, khi_phi_gb
      integer ilo_psi_gb, ihi_psi_gb
      integer jlo_psi_gb, jhi_psi_gb
      integer klo_psi_gb, khi_psi_gb
      integer ilo_fb, ihi_fb, jlo_fb, jhi_fb, klo_fb, khi_fb
      double precision phi(ilo_phi_gb:ihi_phi_gb,
     &                     jlo_phi_gb:jhi_phi_gb,
     &                     klo_phi_gb:khi_phi_gb)
      double precision psi(ilo_psi_gb:ihi_psi_gb,
     &                     jlo_psi_gb:jhi_psi_gb,
     &                     klo_psi_gb:khi_psi_gb)
      double precision xi(1:3)
      double precision xi_normalized(1:3)
      double precision pt1(1:3)
      double precision pt2(1:3)
      double precision x_lower(1:3)
      double precision domain_dimensions(1:3)
      double precision dx(1:3)
      integer i,j,k
      double precision x,y,z
      double precision t_x, t_y, t_z
      double precision n_x, n_y, n_z
      double precision norm
      double precision proj
      double precision normal_proj_domain_dims
      double precision wave_number_normal
      double precision wave_number_tangent
      double precision proj_tangent
      double precision proj_normal
      double precision pi
      parameter (pi=3.14159265358979d0)

c     normalize xi
      norm = sqrt(xi(1)**2 + xi(2)**2 + xi(3)**2)
      xi_normalized(1) = xi(1)/norm
      xi_normalized(2) = xi(2)/norm
      xi_normalized(3) = xi(3)/norm

c     compute the tangent vector in the plane of the dislocation 
c     array that is perpendicular to the direction of the dislocation
c     lines, xi.  also compute the wave number for the sinusoidal 
c     varying level set function values
      t_x = pt1(1) - pt2(1)
      t_y = pt1(2) - pt2(2) 
      t_z = pt1(3) - pt2(3) 
      proj = t_x*xi_normalized(1) + t_y*xi_normalized(2) 
     &     + t_z*xi_normalized(3)  
      t_x = t_x - proj*xi_normalized(1) 
      t_y = t_y - proj*xi_normalized(2) 
      t_z = t_z - proj*xi_normalized(3) 
      norm = sqrt(t_x**2 + t_y**2 + t_z**2)
      wave_number_tangent = pi/norm
      t_x = t_x/norm
      t_y = t_y/norm
      t_z = t_z/norm

      
c     compute the normal vector in the plane of the dislocation 
      n_x = t_y*xi_normalized(3) - t_z*xi_normalized(2)
      n_y = t_z*xi_normalized(1) - t_x*xi_normalized(3)
      n_z = t_x*xi_normalized(2) - t_y*xi_normalized(1)
      norm = sqrt(n_x**2 + n_y**2 + n_z**2)
      n_x = n_x/norm
      n_y = n_y/norm
      n_z = n_z/norm


c     compute wave number in direction normal to dislocation array
      normal_proj_domain_dims = domain_dimensions(1)*n_x 
     &                        + domain_dimensions(2)*n_y
     &                        + domain_dimensions(3)*n_z
      if (normal_proj_domain_dims .lt. 1e-8) then
        normal_proj_domain_dims = domain_dimensions(1)*n_x 
     &                          + domain_dimensions(2)*n_y
     &                          - domain_dimensions(3)*n_z
      endif
      if (normal_proj_domain_dims .lt. 1e-8) then
        normal_proj_domain_dims = domain_dimensions(1)*n_x 
     &                          - domain_dimensions(2)*n_y
     &                          - domain_dimensions(3)*n_z
      endif
      wave_number_normal = abs(pi/normal_proj_domain_dims)

c     loop over grid {
      do k=klo_fb,khi_fb
        do j=jlo_fb,jhi_fb
          do i=ilo_fb,ihi_fb

            x = x_lower(1) + dx(1)*(i-ilo_fb+0.5)
            y = x_lower(2) + dx(2)*(j-jlo_fb+0.5)
            z = x_lower(3) + dx(3)*(k-klo_fb+0.5)

c           compute projection onto plane of dislocation array
            proj_tangent = (x - pt1(1))*t_x + (y - pt1(2))*t_y
     &                   + (z - pt1(3))*t_z

c           compute projection onto normal to plane of dislocation array
            proj_normal = (x - pt1(1))*n_x + (y - pt1(2))*n_y
     &                  + (z - pt1(3))*n_z

c           set phi and psi
            if (cos(wave_number_tangent*proj_tangent) .gt. 0) then
              phi(i,j,k) = sin(wave_number_normal*proj_normal)
            else
              phi(i,j,k) = -sin(wave_number_normal*proj_normal)
            endif
            psi(i,j,k) = sin(wave_number_tangent*proj_tangent)

          enddo
        enddo
      enddo
c     } end loop over grid 

      return
      end
c   } end subroutine
c***********************************************************************
