c***********************************************************************
c
c  File:        lsmdd_utilities.f
c  Copyright:   (c) 2005-2007 Princeton University
c  Author(s):   Kevin T. Chu
c  Revision:    $Revision: 376 $
c  Modified:    $Date: 2007-08-06 19:10:55 -0400 (Mon, 06 Aug 2007) $
c  Description: F77 utility subroutines for level set method 
c               dislocation dynamics
c
c***********************************************************************

c***********************************************************************
c
c  lsmddComputeDislocationLineField() computes the dislocation line 
c  field given the gradient fields for phi and psi.
c
c  The expression for the dislocation line field is: 
c
c    delta(psi) * delta(phi) * unit_tangent_vector
c
c
c  Arguments:
c    disloc_line_* (out):  components of dislocation line field
c    phi (in):             phi
c    grad_phi_* (in):      components of grad(phi)
c    psi (in):             psi
c    grad_psi_* (in):      components of grad(psi)
c    *_gb (in):            index range for ghostbox
c    *_fb (in):            index range for fillbox on which to 
c                          compute dislocation line
c    dX (in):              double array holding grid cell spacing
c                          in the three coordinate directions
c    core_radius (in):     radius of dislocation core 
c
c***********************************************************************
      subroutine lsmddComputeDislocationLineField(
     &  disloc_line_x, disloc_line_y, disloc_line_z,
     &  ilo_disloc_line_gb, ihi_disloc_line_gb,
     &  jlo_disloc_line_gb, jhi_disloc_line_gb,
     &  klo_disloc_line_gb, khi_disloc_line_gb,
     &  phi,
     &  ilo_phi_gb, ihi_phi_gb,
     &  jlo_phi_gb, jhi_phi_gb,
     &  klo_phi_gb, khi_phi_gb,
     &  grad_phi_x, grad_phi_y, grad_phi_z,
     &  ilo_grad_phi_gb, ihi_grad_phi_gb,
     &  jlo_grad_phi_gb, jhi_grad_phi_gb,
     &  klo_grad_phi_gb, khi_grad_phi_gb,
     &  psi,
     &  ilo_psi_gb, ihi_psi_gb,
     &  jlo_psi_gb, jhi_psi_gb,
     &  klo_psi_gb, khi_psi_gb,
     &  grad_psi_x, grad_psi_y, grad_psi_z,
     &  ilo_grad_psi_gb, ihi_grad_psi_gb,
     &  jlo_grad_psi_gb, jhi_grad_psi_gb,
     &  klo_grad_psi_gb, khi_grad_psi_gb,
     &  ilo_fb, ihi_fb,
     &  jlo_fb, jhi_fb,
     &  klo_fb, khi_fb,
     &  dX, core_radius)
c***********************************************************************
c { begin subroutine
      implicit none

c     _gb refers to ghostbox 
c     _fb refers to fillbox on which to compute dislocation line
      integer ilo_disloc_line_gb, ihi_disloc_line_gb
      integer jlo_disloc_line_gb, jhi_disloc_line_gb
      integer klo_disloc_line_gb, khi_disloc_line_gb
      integer ilo_phi_gb, ihi_phi_gb
      integer jlo_phi_gb, jhi_phi_gb
      integer klo_phi_gb, khi_phi_gb
      integer ilo_grad_phi_gb, ihi_grad_phi_gb
      integer jlo_grad_phi_gb, jhi_grad_phi_gb
      integer klo_grad_phi_gb, khi_grad_phi_gb
      integer ilo_psi_gb, ihi_psi_gb
      integer jlo_psi_gb, jhi_psi_gb
      integer klo_psi_gb, khi_psi_gb
      integer ilo_grad_psi_gb, ihi_grad_psi_gb
      integer jlo_grad_psi_gb, jhi_grad_psi_gb
      integer klo_grad_psi_gb, khi_grad_psi_gb
      integer ilo_fb, ihi_fb
      integer jlo_fb, jhi_fb
      integer klo_fb, khi_fb
      double precision disloc_line_x(
     &  ilo_disloc_line_gb:ihi_disloc_line_gb,
     &  jlo_disloc_line_gb:jhi_disloc_line_gb,
     &  klo_disloc_line_gb:khi_disloc_line_gb)
      double precision disloc_line_y(
     &  ilo_disloc_line_gb:ihi_disloc_line_gb,
     &  jlo_disloc_line_gb:jhi_disloc_line_gb,
     &  klo_disloc_line_gb:khi_disloc_line_gb)
      double precision disloc_line_z(
     &  ilo_disloc_line_gb:ihi_disloc_line_gb,
     &  jlo_disloc_line_gb:jhi_disloc_line_gb,
     &  klo_disloc_line_gb:khi_disloc_line_gb)
      double precision phi(ilo_phi_gb:ihi_phi_gb,
     &                     jlo_phi_gb:jhi_phi_gb,
     &                     klo_phi_gb:khi_phi_gb)
      double precision grad_phi_x(ilo_grad_phi_gb:ihi_grad_phi_gb,
     &                            jlo_grad_phi_gb:jhi_grad_phi_gb,
     &                            klo_grad_phi_gb:khi_grad_phi_gb)
      double precision grad_phi_y(ilo_grad_phi_gb:ihi_grad_phi_gb,
     &                            jlo_grad_phi_gb:jhi_grad_phi_gb,
     &                            klo_grad_phi_gb:khi_grad_phi_gb)
      double precision grad_phi_z(ilo_grad_phi_gb:ihi_grad_phi_gb,
     &                            jlo_grad_phi_gb:jhi_grad_phi_gb,
     &                            klo_grad_phi_gb:khi_grad_phi_gb)
      double precision psi(ilo_psi_gb:ihi_psi_gb,
     &                     jlo_psi_gb:jhi_psi_gb,
     &                     klo_psi_gb:khi_psi_gb)
      double precision grad_psi_x(ilo_grad_psi_gb:ihi_grad_psi_gb,
     &                            jlo_grad_psi_gb:jhi_grad_psi_gb,
     &                            klo_grad_psi_gb:khi_grad_psi_gb)
      double precision grad_psi_y(ilo_grad_psi_gb:ihi_grad_psi_gb,
     &                            jlo_grad_psi_gb:jhi_grad_psi_gb,
     &                            klo_grad_psi_gb:khi_grad_psi_gb)
      double precision grad_psi_z(ilo_grad_psi_gb:ihi_grad_psi_gb,
     &                            jlo_grad_psi_gb:jhi_grad_psi_gb,
     &                            klo_grad_psi_gb:khi_grad_psi_gb)
      double precision dX(1:3)
      double precision core_radius
      integer i,j,k
      double precision inv_core_radius
      double precision delta_phi, delta_psi
      double precision norm_tangent_vector, inv_norm_tangent_vector
      double precision pi
      parameter (pi=3.14159265358979323846d0)
      double precision zero_tol
      parameter (zero_tol=1.0d-13)

c     compute inv_core_radius
      inv_core_radius = 1.0d0/core_radius

c     loop over fillbox {
      do k=klo_fb,khi_fb
        do j=jlo_fb,jhi_fb
          do i=ilo_fb,ihi_fb

c           compute tangent vector (extended off of dislocation line 
c           by using the formula:
c           
c             grad(phi) cross grad(psi) / |grad(phi) cross grad(psi)|
c           
            disloc_line_x(i,j,k) = grad_phi_y(i,j,k)*grad_psi_z(i,j,k)
     &                           - grad_phi_z(i,j,k)*grad_psi_y(i,j,k)
            disloc_line_y(i,j,k) = grad_phi_z(i,j,k)*grad_psi_x(i,j,k)
     &                           - grad_phi_x(i,j,k)*grad_psi_z(i,j,k)
            disloc_line_z(i,j,k) = grad_phi_x(i,j,k)*grad_psi_y(i,j,k)
     &                           - grad_phi_y(i,j,k)*grad_psi_x(i,j,k)

c           compute norm_tangent_vector
            norm_tangent_vector = sqrt(
     &          disloc_line_x(i,j,k)*disloc_line_x(i,j,k)
     &        + disloc_line_y(i,j,k)*disloc_line_y(i,j,k)
     &        + disloc_line_z(i,j,k)*disloc_line_z(i,j,k) )
            
            if (norm_tangent_vector .ge. zero_tol) then

c             compute smoothed delta functions at dislocation line  
              if (abs(phi(i,j,k)) .gt. core_radius) then
                delta_phi = 0.d0
              else
                delta_phi = 0.5d0*inv_core_radius
     &                    * ( 1+cos(pi*phi(i,j,k)*inv_core_radius) )
              endif
              if (abs(psi(i,j,k)) .gt. core_radius) then
                delta_psi = 0.d0
              else
                delta_psi = 0.5d0*inv_core_radius
     &                    * ( 1+cos(pi*psi(i,j,k)*inv_core_radius) )
              endif

c             calculate dislocation line fields by normalizing
c             and multiplying by delta function

              inv_norm_tangent_vector = 1.0d0/norm_tangent_vector 
              disloc_line_x(i,j,k) = disloc_line_x(i,j,k) 
     &                             * inv_norm_tangent_vector
     &                             * delta_phi * delta_psi
              disloc_line_y(i,j,k) = disloc_line_y(i,j,k) 
     &                             * inv_norm_tangent_vector
     &                             * delta_phi * delta_psi
              disloc_line_z(i,j,k) = disloc_line_z(i,j,k) 
     &                             * inv_norm_tangent_vector
     &                             * delta_phi * delta_psi

            else

c             tangent vector is basically zero
              disloc_line_x(i,j,k) = 0.0d0
              disloc_line_y(i,j,k) = 0.0d0
              disloc_line_z(i,j,k) = 0.0d0

            endif

          enddo
        enddo
      enddo
c     } end loop over grid 

      return
      end
c } end subroutine
c***********************************************************************

c***********************************************************************
c
c  lsmddComputeUnitTangentVectorField() computes the unit tangent 
c  vector field for a dislocation line as the cross product of the 
c  gradient fields for phi and psi.  That is, 
c
c    unit_tangent_vector = grad(phi) cross grad(psi) 
c                        / | grad(phi) cross grad(psi) |
c
c  Arguments:
c    unit_tangent_* (out):  components of dislocation line field
c    grad_phi_* (in):       components of grad(phi)
c    grad_psi_* (in):       components of grad(psi)
c    *_gb (in):             index range for ghostbox
c    *_fb (in):             index range for fillbox on which to 
c                           compute dislocation line
c    dX (in):               double array holding grid cell spacing
c                           in the three coordinate directions
c
c***********************************************************************
      subroutine lsmddComputeUnitTangentVectorField(
     &  unit_tangent_x, unit_tangent_y, unit_tangent_z,
     &  ilo_unit_tangent_gb, ihi_unit_tangent_gb,
     &  jlo_unit_tangent_gb, jhi_unit_tangent_gb,
     &  klo_unit_tangent_gb, khi_unit_tangent_gb,
     &  grad_phi_x, grad_phi_y, grad_phi_z,
     &  ilo_grad_phi_gb, ihi_grad_phi_gb,
     &  jlo_grad_phi_gb, jhi_grad_phi_gb,
     &  klo_grad_phi_gb, khi_grad_phi_gb,
     &  grad_psi_x, grad_psi_y, grad_psi_z,
     &  ilo_grad_psi_gb, ihi_grad_psi_gb,
     &  jlo_grad_psi_gb, jhi_grad_psi_gb,
     &  klo_grad_psi_gb, khi_grad_psi_gb,
     &  ilo_fb, ihi_fb,
     &  jlo_fb, jhi_fb,
     &  klo_fb, khi_fb,
     &  dX)
c***********************************************************************
c { begin subroutine
      implicit none

c     _gb refers to ghostbox 
c     _fb refers to fillbox on which to compute unit tangent vector
      integer ilo_unit_tangent_gb, ihi_unit_tangent_gb
      integer jlo_unit_tangent_gb, jhi_unit_tangent_gb
      integer klo_unit_tangent_gb, khi_unit_tangent_gb
      integer ilo_grad_phi_gb, ihi_grad_phi_gb
      integer jlo_grad_phi_gb, jhi_grad_phi_gb
      integer klo_grad_phi_gb, khi_grad_phi_gb
      integer ilo_grad_psi_gb, ihi_grad_psi_gb
      integer jlo_grad_psi_gb, jhi_grad_psi_gb
      integer klo_grad_psi_gb, khi_grad_psi_gb
      integer ilo_fb, ihi_fb
      integer jlo_fb, jhi_fb
      integer klo_fb, khi_fb
      double precision unit_tangent_x(
     &  ilo_unit_tangent_gb:ihi_unit_tangent_gb,
     &  jlo_unit_tangent_gb:jhi_unit_tangent_gb,
     &  klo_unit_tangent_gb:khi_unit_tangent_gb)
      double precision unit_tangent_y(
     &  ilo_unit_tangent_gb:ihi_unit_tangent_gb,
     &  jlo_unit_tangent_gb:jhi_unit_tangent_gb,
     &  klo_unit_tangent_gb:khi_unit_tangent_gb)
      double precision unit_tangent_z(
     &  ilo_unit_tangent_gb:ihi_unit_tangent_gb,
     &  jlo_unit_tangent_gb:jhi_unit_tangent_gb,
     &  klo_unit_tangent_gb:khi_unit_tangent_gb)
      double precision grad_phi_x(ilo_grad_phi_gb:ihi_grad_phi_gb,
     &                            jlo_grad_phi_gb:jhi_grad_phi_gb,
     &                            klo_grad_phi_gb:khi_grad_phi_gb)
      double precision grad_phi_y(ilo_grad_phi_gb:ihi_grad_phi_gb,
     &                            jlo_grad_phi_gb:jhi_grad_phi_gb,
     &                            klo_grad_phi_gb:khi_grad_phi_gb)
      double precision grad_phi_z(ilo_grad_phi_gb:ihi_grad_phi_gb,
     &                            jlo_grad_phi_gb:jhi_grad_phi_gb,
     &                            klo_grad_phi_gb:khi_grad_phi_gb)
      double precision grad_psi_x(ilo_grad_psi_gb:ihi_grad_psi_gb,
     &                            jlo_grad_psi_gb:jhi_grad_psi_gb,
     &                            klo_grad_psi_gb:khi_grad_psi_gb)
      double precision grad_psi_y(ilo_grad_psi_gb:ihi_grad_psi_gb,
     &                            jlo_grad_psi_gb:jhi_grad_psi_gb,
     &                            klo_grad_psi_gb:khi_grad_psi_gb)
      double precision grad_psi_z(ilo_grad_psi_gb:ihi_grad_psi_gb,
     &                            jlo_grad_psi_gb:jhi_grad_psi_gb,
     &                            klo_grad_psi_gb:khi_grad_psi_gb)
      double precision dX(1:3)
      integer i,j,k
      double precision norm_tangent_vector, inv_norm_tangent_vector
      double precision pi
      parameter (pi=3.14159265358979323846d0)
      double precision zero_tol
      parameter (zero_tol=1.0d-13)


c     loop over fillbox {
      do k=klo_fb,khi_fb
        do j=jlo_fb,jhi_fb
          do i=ilo_fb,ihi_fb

c           compute tangent vector (extended off of dislocation line 
c           by using the formula:
c           
c             grad(phi) cross grad(psi) / |grad(phi) cross grad(psi)|
c           
            unit_tangent_x(i,j,k) = grad_phi_y(i,j,k)*grad_psi_z(i,j,k)
     &                            - grad_phi_z(i,j,k)*grad_psi_y(i,j,k)
            unit_tangent_y(i,j,k) = grad_phi_z(i,j,k)*grad_psi_x(i,j,k)
     &                            - grad_phi_x(i,j,k)*grad_psi_z(i,j,k)
            unit_tangent_z(i,j,k) = grad_phi_x(i,j,k)*grad_psi_y(i,j,k)
     &                            - grad_phi_y(i,j,k)*grad_psi_x(i,j,k)

c           compute norm of tangent vector
            norm_tangent_vector = sqrt(
     &          unit_tangent_x(i,j,k)*unit_tangent_x(i,j,k)
     &        + unit_tangent_y(i,j,k)*unit_tangent_y(i,j,k)
     &        + unit_tangent_z(i,j,k)*unit_tangent_z(i,j,k) )
            
c           normalize tangent vector 
            if (norm_tangent_vector .ge. zero_tol) then

              inv_norm_tangent_vector = 1.0d0/norm_tangent_vector 
              unit_tangent_x(i,j,k) = unit_tangent_x(i,j,k) 
     &                              * inv_norm_tangent_vector
              unit_tangent_y(i,j,k) = unit_tangent_y(i,j,k) 
     &                              * inv_norm_tangent_vector
              unit_tangent_z(i,j,k) = unit_tangent_z(i,j,k) 
     &                              * inv_norm_tangent_vector
            else

c             norm_tangent_vector is too small, so set tangent arbitrarily
              unit_tangent_x(i,j,k) = 1.0d0
              unit_tangent_y(i,j,k) = 0.0d0
              unit_tangent_z(i,j,k) = 0.0d0

            endif

          enddo
        enddo
      enddo
c     } end loop over grid 

      return
      end
c } end subroutine
c***********************************************************************
